# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from feijoa import create_job
from feijoa.search.algorithms.bayesian import BayesianAlgorithm
from feijoa.search.algorithms.genetic import CMAES
from feijoa.search.algorithms.genetic import PSO
from feijoa.search.algorithms.pattern import PatternSearch
from feijoa.search.algorithms.randomized import RandomSearch
from feijoa.search.bandit import ThompsonSampler
from feijoa.search.parameters import Categorical
from feijoa.search.parameters import Real
from feijoa.search.space import SearchSpace
from feijoa.utils.logging import init


def pad(string: str, width: int, filler=" ", fill_chars=3):
    """Pads a string to a given width.

    Args:
        string (str):
            String instance.
        width (int):
            Target width of string.
        filler:
            Fill places with specified char.
        fill_chars:
            Count of chars to fill.

    Raises:
        AnyError: If anything bad happens.

    """

    if len(string) <= width:
        return string
    return f"{string[:width - 3]}{filler * fill_chars}"


def plot_evaluations(job, params=None, invert_objective=False):
    """Plot evaluations of job.

    Args:
        job (Job):
            Job instance.
        params (list):
            List of parameters to plot.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    df = job.get_dataframe(brief=True, only_good=True)
    objectives = df["objective_result"] * (
        1 if not invert_objective else -1
    )
    df.drop(columns=["id", "objective_result"], inplace=True)
    params = params or df.columns

    if any([param not in df.columns for param in params]):
        raise ValueError("Unknown columns in specified job.")

    params_length = len(params)

    fig = make_subplots(rows=params_length, cols=params_length)

    for i, col_x in enumerate(params):
        for j, col_y in enumerate(params):
            if i > j:
                continue

            if i == j:
                fig.add_histogram(
                    x=df[col_x],
                    nbinsx=30,
                    row=i + 1,
                    col=i + 1,
                    marker=dict(
                        color="#330C73",
                    ),
                    opacity=0.75,
                )
                continue

            x = df[col_x]
            y = df[col_y]

            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode="markers",
                    marker=dict(
                        color=objectives,
                        colorscale="TealRose_r",
                        symbol="circle",
                        colorbar=dict(title="Objective result"),
                        line=dict(width=1, color="black"),
                    ),
                ),
                col=i + 1,
                row=j + 1,
            )

    fig.update_layout(
        title="Evaluations plot",
        showlegend=False,
        bargap=0.2,
        bargroupgap=0.1,
    )

    def cut(base, num):
        if num == 1:
            return base
        return base if num == 1 else f"{base}{num}"

    positions = [
        (
            *fig["layout"][cut("xaxis", i)]["domain"],
            *fig["layout"][cut("yaxis", i)]["domain"],
        )
        for i in range(1, params_length**2 + 1)
    ]

    positions = [
        (x0, x1, y0, y1)
        for (x0, x1, y0, y1) in positions
        if x0 == 0.0 or y0 == 0.0
    ]

    names = list(params) * 2

    for i, (x0, x1, y0, y1) in enumerate(positions):
        x = x0
        y = y0
        angle = 0

        if x == 0.0:
            x -= 0.05
            y = (y0 + y1) / 2
            angle = 270
        else:
            x = (x0 + x1) / 2 + 0.02
            y -= 0.08

        fig.add_annotation(
            dict(
                x=x,
                y=y,
                xref="paper",
                yref="paper",
                text=pad(names.pop(0), 10, filler="."),
                showarrow=False,
            ),
            textangle=angle,
        )

        if x0 == 0 and y0 == 0:
            x_ = (x0 + x1) / 2
            y_ = -0.08
            angle = 0
            fig.add_annotation(
                dict(
                    x=x_,
                    y=y_,
                    xref="paper",
                    yref="paper",
                    text=pad(names.pop(0), 10, filler="."),
                    showarrow=False,
                ),
                textangle=angle,
            )

    return fig


def objective(experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    z = experiment.params.get("z")
    w = experiment.params.get("w")
    fast = experiment.params.get("fast")
    u = (
        (
            (1.5 - x + x * y) ** 2
            + (2.25 - x + x * y**2) ** 2
            + (2.625 - x + x * y**3) ** 2
        )
        - 0.1 * z
        - 0.5 * w
    )

    if fast == "foo":
        u -= 10
    elif fast == "bar":
        pass
    else:
        u -= 5

    return u


def main():
    space = SearchSpace()

    init(verbose=True)

    space.insert(Real("x", low=0.0, high=3.0))
    space.insert(Real("y", low=0.0, high=1.0))
    space.insert(Real("z", low=0.0, high=1.0))
    space.insert(Real("w", low=0.0, high=1.0))
    space.insert(Categorical("fast", choices=["foo", "bar", "zoo"]))

    bayesian = BayesianAlgorithm(search_space=space)
    template = PatternSearch(search_space=space)
    cmaes = CMAES(search_space=space)
    pso_ = PSO(search_space=space)
    rnd = RandomSearch(search_space=space)

    thompson = ThompsonSampler(bayesian, template, rnd, cmaes, pso_)

    job = create_job(search_space=space)
    job.do(
        objective,
        n_jobs=-1,
        n_trials=100,
        algo_list=[thompson],
        progress_bar=True,
        use_numba_jit=True,
    )

    fig = plot_evaluations(job)
    fig.show()


if __name__ == "__main__":
    main()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PCA importance evaluator module."""

import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import LabelEncoder

from ..jobs.job import Job
from .evaluator import ImportanceEvaluator


__all__ = ["PCAEvaluator"]


# noinspection DuplicatedCode
class PCAEvaluator(ImportanceEvaluator):
    """Principal component analysis (PCA) importance evaluator.

    .. code-block:: python

        from feijoa.importance.pca import PCAEvaluator

        job = ...
        evaluator = PCAEvaluator()
        imp = evaluator.do(job)

        params = imp["params"]
        importances = imp["importances"]
    """

    def __init__(self):
        self.pca = PCA()

    def do(self, job: Job):
        df = job.get_dataframe(brief=True, only_good=True)
        y = df["objective_result"]
        X = df.drop(columns=["objective_result", "id"])
        categorical = X.select_dtypes(include=["category"])
        encoder = LabelEncoder()
        for cat in categorical.columns:
            X[cat] = encoder.fit_transform(X[cat])

        self.pca.fit(X, y)
        importance = np.array(self.pca.explained_variance_)
        completed = dict()
        completed["parameters"] = X.columns
        completed["importance"] = np.array(importance)
        return completed



import sys


def in_notebook():
    """Check if module in running in IPython kernel.

    Returns:
         ``True`` if the module is running in IPython kernel,
         ``False`` if in IPython shell or other Python shell.

    Raises:
        AnyError: If anything bad happens.

    """
    return "ipykernel" in sys.modules



import pytest

from feijoa.search import SearchSpace

# noinspection PyProtectedMember,PyDeprecation
from feijoa.search.space import from_yaml


def test_space_from_yaml():
    yaml = """- signature: -O
  type: categorical
  choices: [-O1, -O2, -O3, null]

- signature: align-functions
  type: categorical
  choices: [-falign-functions, -fno-align-functions, null]

- signature: iv-max-considered-uses
  type: integer
  range: [0, 1000]

- signature: foo
  type: integer
  low: 0
  high: 1000

- signature: boo
  type: real
  low: 0.0
  high: 1.0

- signature: zoo
  type: real
  range: [0.0, 10.0]
"""

    space = SearchSpace.from_yaml(yaml)

    assert space.get("boo").name == "boo"

    with open("foo.yaml", "w") as yf:
        yf.write(yaml)

    with pytest.deprecated_call():
        # noinspection PyDeprecation
        from_yaml("foo.yaml")

    space_from_file = SearchSpace.from_yaml_file("foo.yaml")

    assert (
        str(space)
        == """SearchSpace:
    Categorical('-O', choices=['-O1', '-O2', '-O3', None])
    Categorical('align-functions', choices=['-falign-functions', '-fno-align-functions', None])
    Integer('iv-max-considered-uses, low=0, high=1000)
    Integer('foo, low=0, high=1000)
    Real('boo', low=0.0, high=1.0)
    Real('zoo', low=0.0, high=10.0)"""
    )

    assert (
        str(space_from_file)
        == """SearchSpace:
    Categorical('-O', choices=['-O1', '-O2', '-O3', None])
    Categorical('align-functions', choices=['-falign-functions', '-fno-align-functions', None])
    Integer('iv-max-considered-uses, low=0, high=1000)
    Integer('foo, low=0, high=1000)
    Real('boo', low=0.0, high=1.0)
    Real('zoo', low=0.0, high=10.0)"""
    )


def test_incorrect_yaml_parameter_type():
    doc = """- signature: foo
  type: unknown
  low: 0
  high: 1000"""

    with pytest.raises(TypeError):
        SearchSpace.from_yaml(doc)



from feijoa import Categorical
from feijoa import create_job
from feijoa import Integer
from feijoa import Real
from feijoa import SearchSpace
from feijoa.search.algorithms.bayesian import BayesianAlgorithm


def test_bayesian_search():
    space = SearchSpace()

    space.insert(Integer("x", low=0, high=2))
    space.insert(Real("y", low=0.0, high=0.5))
    space.insert(Categorical("z", choices=["foo", "bar"]))

    def objective(experiment):
        x = experiment.params.get("x")
        y = experiment.params.get("y")
        z = experiment.params.get("z")

        return -(x + y + (2 if z == "foo" else 3))

    job = create_job(search_space=space)
    job.do(objective, n_trials=50, algo_list=["bayesian"])


# noinspection DuplicatedCode
def test_specify_acq_bayesian_search():
    space = SearchSpace()

    space.insert(Integer("x", low=0, high=2))
    space.insert(Real("y", low=0.0, high=0.5))
    space.insert(Categorical("z", choices=["foo", "bar"]))

    def objective(experiment):
        x = experiment.params.get("x")
        y = experiment.params.get("y")
        z = experiment.params.get("z")

        return -(x + y + (2 if z == "foo" else 3))

    job = create_job(search_space=space)

    a_ei = BayesianAlgorithm(
        search_space=job.search_space, acq_function="ei"
    )

    job.do(objective, n_trials=10, algo_list=[a_ei])

    a_poi = BayesianAlgorithm(
        search_space=job.search_space, acq_function="poi"
    )

    job.do(objective, n_trials=10, algo_list=[a_poi])

    a_ucb = BayesianAlgorithm(
        search_space=job.search_space, acq_function="ucb"
    )

    job.do(objective, n_trials=10, algo_list=[a_ucb])



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""SQLAlchemy models module."""

import json

from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import TypeDecorator
from sqlalchemy import types
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import relationship


__all__ = [
    "SearchSpaceModel",
    "ParameterModel",
    "JobModel",
    "ExperimentModel",
]


_Base: DeclarativeMeta = declarative_base()


class Json(TypeDecorator):
    """Json column (temporary class)."""

    @property
    def python_type(self):
        return object

    impl = types.String

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_literal_param(self, value, dialect):
        return value

    def process_result_value(self, value, dialect):
        try:
            return json.loads(value)
        except (ValueError, TypeError):
            return None


class SearchSpaceModel(_Base):
    __tablename__ = "search_space"

    id = Column(Integer, primary_key=True)


class ParameterModel(_Base):
    __tablename__ = "parameter"

    id = Column(Integer, primary_key=True)
    search_space_id = Column(ForeignKey(SearchSpaceModel.id))
    search_space = relationship(
        SearchSpaceModel, backref="parameters"
    )

    name = Column(String)
    kind = Column(String)
    meta = Column(Json)


class JobModel(_Base):
    __tablename__ = "jobs"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    search_space_id = Column(ForeignKey(SearchSpaceModel.id))
    search_space = relationship(SearchSpaceModel, backref="jobs")

    def __repr__(self):
        return f"Job<{self.id}, {self.name}"


class ExperimentModel(_Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True)
    job_id = Column(ForeignKey(JobModel.id), primary_key=True)
    job = relationship(JobModel, backref="experiments")
    state = Column(String)
    hash = Column(String)
    objective_result = Column(Float)
    requestor = Column(String)
    params = Column(Json)
    create_timestamp = Column(Float)
    finish_timestamp = Column(Float)
    metrics = Column(Json)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Generator
from typing import List
from typing import Optional

from feijoa.models.configuration import Configuration
from feijoa.search.algorithms import SearchAlgorithm
from feijoa.search.visitors import Randomizer


__all__ = ["RandomSearch"]


class RandomSearch(SearchAlgorithm):
    """Simple random search.

    Random search replaces the exhaustive
    search of all combinations with a
    selection of them randomly. This can
    easily be applied to the discrete settings
    above, but the method can also be
    generalized to continuous and mixed spaces.
    Random search can outperform lattice search,
    especially if only a small number of hyperparameters
    affects the performance of the algorithm.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "random"
    aliases = ("RandomSearch", "random", "randomized")

    def __init__(self, search_space, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_space = search_space
        self.randomizer = Randomizer()
        self._ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self._ask_gen:
            self._ask_gen = self._ask(n)
        return next(self._ask_gen)

    def _ask(self, n: int) -> Generator:

        while True:
            cfgs = []

            for _ in range(n):
                cfg = dict()

                # random sample

                for p in self.search_space:
                    cfg[p.name] = p.accept(self.randomizer)

                cfgs.append(Configuration(cfg, requestor=self.name))

            yield cfgs

    def tell(self, config, result):
        """No needed."""

        pass



import hashlib
import json

from feijoa import Experiment
from feijoa.models.experiment import ExperimentState


def test_experiment():

    ex1 = Experiment(
        id=0,
        job_id=0,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params={"x": 0.0, "y": 1.0},
    )

    print(str(ex1))

    assert (
        str(ex1)
        == """Experiment({
    "id": 0,
    "job_id": 0,
    "state": "WIP",
    "hash": null,
    "objective_result": null,
    "params": {
        "x": 0.0,
        "y": 1.0
    },
    "create_timestamp": 0.0,
    "finish_timestamp": null,
    "metrics": null
})"""
    )

    assert not ex1.is_finished()

    ex1.apply(1.0)
    ex1.success_finish()

    ex1_hash = ex1.hash
    hash_len = len(ex1_hash)

    params_part = ex1_hash[: hash_len // 2]
    objective_part = ex1_hash[hash_len // 2 :]

    params_dumped = json.dumps(ex1.params, sort_keys=True)

    assert (
        hashlib.sha1(params_dumped.encode()).hexdigest()
        == params_part
    )
    assert (
        hashlib.sha1(str(ex1.objective_result).encode()).hexdigest()
        == objective_part
    )

    ex2 = Experiment(
        id=0,
        job_id=0,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params={"x": 0.0, "y": 1.0},
    )

    ex2.set_error()
    assert ex2.state == ExperimentState.ERROR

    ex2.error_finish()
    assert ex2.finish_timestamp



import glob
from itertools import chain
import logging
import os

import pytest


excluded_loggers = (
    "numba",
    "matplotlib",
)

for log_name in excluded_loggers:
    other_log = logging.getLogger(log_name)
    other_log.setLevel(logging.WARNING)


@pytest.fixture(autouse=True)
def cleanup():
    yield
    for f in chain(
        glob.glob("*.json"), glob.glob("*.yaml"), glob.glob("*.db")
    ):
        os.remove(f)



from feijoa import create_job
from feijoa import Experiment
from feijoa import Real
from feijoa import SearchSpace


# noinspection DuplicatedCode
def test_seed():
    space = SearchSpace()
    space.insert(Real("x", low=0.0, high=1.0))
    space.insert(Real("y", low=0.0, high=1.0))

    def objective(experiment: Experiment):
        params = experiment.params

        x = params.get("x", 5.0)
        y = params.get("y", 5.0)

        return (1 - x) ** 2 + (1 - y) ** 2

    job = create_job(search_space=space)
    job.add_seed({"x": 1.0, "y": 1.0})
    job.do(objective, n_trials=50, n_jobs=1, algo_list=["grid"])

    assert job.best_parameters == {"x": 1.0, "y": 1.0}
    # assert job.best_experiment.requestor == "SeedAlgorithm"



from feijoa import Categorical
from feijoa import create_job
from feijoa import Experiment
from feijoa import Integer
from feijoa import Real
from feijoa import SearchSpace
from feijoa.utils.imports import import_or_skip


import_or_skip("skopt")


# noinspection DuplicatedCode
def objective(experiment: Experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    return (
        (1.5 - x + x * y) ** 2
        + (2.25 - x + x * y**2) ** 2
        + (2.625 - x + x * y**3) ** 2
    )


def test_skopt():
    # noinspection DuplicatedCode
    space = SearchSpace()

    space.insert(Real("x", low=0.0, high=5.0))
    space.insert(Real("y", low=0.0, high=2.0))
    space.insert(Integer("z", low=0, high=1))
    space.insert(Categorical("w", choices=["foo"]))

    job = create_job(search_space=space)
    job.do(objective, n_trials=10, algo_list=["skopt"])



from sklearn.ensemble import AdaBoostRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.neural_network import MLPRegressor

from feijoa import create_job
from feijoa import Experiment
from feijoa import Real
from feijoa import SearchSpace

# noinspection DuplicatedCode
from feijoa.search.algorithms.bayesian import BayesianAlgorithm
from feijoa.search.bandit import ThompsonSampler


def objective(experiment: Experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    return (
        (1.5 - x + x * y) ** 2
        + (2.25 - x + x * y**2) ** 2
        + (2.625 - x + x * y**3) ** 2
    )


def test_minimal():
    space = SearchSpace()

    space.insert(Real("x", low=0.0, high=5.0))
    space.insert(Real("y", low=0.0, high=2.0))

    job = create_job(search_space=space)
    job.do(objective, n_trials=50)

    assert abs(job.best_value - 0) < 5
    assert len(job.dataframe) == 50


def test_job_with_algo_ensemble():
    space = SearchSpace()

    space.insert(Real("x", low=0.0, high=5.0))
    space.insert(Real("y", low=0.0, high=2.0))

    job = create_job(search_space=space, optimizer=ThompsonSampler)

    ada = BayesianAlgorithm(
        search_space=space, regressor=AdaBoostRegressor
    )
    forest = BayesianAlgorithm(
        search_space=space, regressor=RandomForestRegressor
    )
    classic = BayesianAlgorithm(search_space=space)

    classic.name = "ClassicBayesianSearch"
    mlp = BayesianAlgorithm(
        search_space=space, regressor=MLPRegressor
    )

    job.do(
        objective,
        n_trials=50,
        algo_list=["template", ada, forest, classic, mlp],
    )

    assert abs(job.best_value - 0) < 5
    assert len(job.dataframe) == 50



import logging

import numpy as np
import pytest

from feijoa import create_job
from feijoa import Real
from feijoa import SearchSpace
from feijoa.exceptions import InvalidOptimizer
from feijoa.search import RoundRobinMeta
from feijoa.search.algorithms.grid import GridSearch
from feijoa.search.algorithms.pattern import PatternSearch
from feijoa.search.bandit import ThompsonSampler
from feijoa.search.bandit import UCB1
from feijoa.search.bandit import UCBTuned


log = logging.getLogger(__name__)


def test_different_optimizers():
    def f(experiment):
        x, y = experiment.params["x"], experiment.params["y"]
        return x + np.cos(y)

    space = SearchSpace(
        Real("x", low=0.0, high=1.0), Real("y", low=0.0, high=1.0)
    )

    create_job(search_space=space, optimizer=RoundRobinMeta)
    create_job(search_space=space, optimizer=UCB1)

    with pytest.raises(InvalidOptimizer):
        create_job(search_space=space, optimizer=1)

    grid = GridSearch(search_space=space)
    template = PatternSearch(search_space=space)
    template1 = PatternSearch(search_space=space)

    op1 = RoundRobinMeta(grid, template)
    op2 = UCB1(template, grid, template1)
    op3 = UCBTuned(grid, template)

    ensemble = ThompsonSampler(op1, op2, op3)

    job = create_job(search_space=space, optimizer=ensemble)
    job.do(objective=f, n_jobs=5, n_trials=20, clear=False)

    print(job.best_parameters)
    print(job.rewards)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Experiment model class module."""

import datetime
from enum import Enum
import hashlib
import json
from typing import Any
from typing import Optional

from pydantic import BaseModel

from feijoa.models.configuration import Configuration


__all__ = ["ExperimentState", "Experiment"]


class ExperimentState(str, Enum):

    OK = "OK"
    WIP = "WIP"
    ERROR = "ERROR"

    def __repr__(self):
        return self.name

    def __str__(self):
        return repr(self)


# noinspection PyUnresolvedReferences
class Experiment(BaseModel):
    """Experiment model.

    Lifecycle of experiment:

    1) Create experiment from params (state=WIP)
    2) Suggest experiment to job
    3) Measure it
    4) If all metrics are collected set OK state,
       if error caused - ERROR.
    5) Finish experiment - calculate hash, set finish
       timestamp
    6) Tell to optimizers
    6) Save to storage

    Args:
        id (int):
            Index of experiment.
        job_id (int):
            Job index.
        state (ExperimentState):
            Experiment state. Must be `WIP`, `OK` or `ERROR`
        hash (str, optional):
            Hash of experiment.
        objective_result (Optional[Any]):
            Objective result of experiment. Now
            is float only, but in next version
            can be tuple/array.
        create_timestamp (float):
            Experiment creation timestamp.
        finish_timestamp (float):
            Experiment finish timestamp.
        metrics (dict, optional):
            Metrics of experiment.

    Raises:
        AnyError: If anything bad happens.

    """

    id: int
    job_id: int
    state: ExperimentState
    hash: Optional[str]
    objective_result: Optional[Any]
    params: Configuration
    create_timestamp: float
    finish_timestamp: Optional[float]
    metrics: Optional[dict]

    class Config:
        orm_mode = True

    def apply(self, result):
        """Apply result to experiment.

        Args:
            result (float):
                Objective result.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        # TODO (qnbhd): make safe operations
        self.objective_result = result

    def set_error(self):
        """Set error state to experiment."""

        self.state = ExperimentState.ERROR

    def is_finished(self):
        """Check if experiment is finished."""

        return (
            self.state == ExperimentState.OK
            or self.state == ExperimentState.ERROR
        )

    def _calculate_hash(self):
        """Calculate hash."""

        params_dumped = json.dumps(self.params, sort_keys=True)

        curve1 = hashlib.sha1(params_dumped.encode())
        curve2 = hashlib.sha1(str(self.objective_result).encode())

        hd1 = curve1.hexdigest()
        hd2 = curve2.hexdigest()

        return hd1 + hd2

    def _finish(self, state):
        """Finish experiment."""

        self.state = state
        self.hash = self._calculate_hash()

        time = datetime.datetime.now()
        self.finish_timestamp = datetime.datetime.timestamp(time)

    def success_finish(self):
        """Finish experiment with success state."""

        self._finish(ExperimentState.OK)

    def error_finish(self):
        """Finish experiment with error state."""

        self._finish(ExperimentState.ERROR)

    def __repr__(self):
        return f"Experiment({json.dumps(self.dict(), indent=4)})"

    def __str__(self):
        return repr(self)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Scikit-optimize integration module."""

from typing import Generator
from typing import List
from typing import Optional

from numpy import float64
from numpy import int64
from numpy.ma import MaskedArray
import sklearn.utils.fixes

from feijoa.models.configuration import Configuration
from feijoa.search.algorithms import SearchAlgorithm
from feijoa.search.parameters import Categorical
from feijoa.search.parameters import Integer
from feijoa.search.parameters import Real
from feijoa.search.space import SearchSpace
from feijoa.utils.imports import ImportWrapper


sklearn.utils.fixes.MaskedArray = MaskedArray
# noinspection PyPackageRequirements
with ImportWrapper():
    import skopt

__all__ = ["SkoptBayesianAlgorithm"]


class SkoptBayesianAlgorithm(SearchAlgorithm):
    """Bayesian algorithm implementation with
    `scikit-optimize` backend.

    Link: https://scikit-optimize.github.io/.

    Args:
        search_space (SearchSpace):
            Search space instance.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "skopt"
    aliases = ("Skopt", "SkoptBayesian", "skopt")

    def __init__(self, search_space: SearchSpace, **kwargs):

        super().__init__(**kwargs)

        self.skopt_space = self._make_space(search_space)
        self.optimizer_instance = skopt.Optimizer(
            self.skopt_space, "GBRT"
        )

        self.ask_generator: Optional[Generator] = None

    @staticmethod
    def _make_space(space: SearchSpace):
        params = []

        for p in space.params:
            if isinstance(p, Integer):
                params.append(
                    skopt.space.Integer(
                        name=p.name, low=p.low, high=p.high
                    )
                )
            elif isinstance(p, Real):
                params.append(
                    skopt.space.Real(
                        name=p.name, low=p.low, high=p.high
                    )
                )
            elif isinstance(p, Categorical):
                params.append(
                    skopt.space.Categorical(
                        name=p.name, categories=p.choices
                    )
                )

        return params

    def _ask(self, n: int) -> Generator:
        @skopt.utils.use_named_args(self.skopt_space)
        def named(**kwargs) -> dict:
            """A namedtuple.

            (Generated by docly)
            """
            return kwargs

        while True:
            asked = list()
            raws = self.optimizer_instance.ask(n_points=n)

            for raw in raws:
                # FIXME (qnbhd) make correct work with numpy types
                potentially_cfg = dict()

                for k, v in named(raw).items():
                    if isinstance(v, int64):  # type: ignore
                        v = int(v)
                    if isinstance(v, (float, float64)):  # type: ignore
                        v = float(v)

                    potentially_cfg[k] = v

                asked.append(
                    Configuration(
                        potentially_cfg, requestor=self.name
                    )
                )

            yield asked

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self.ask_generator:
            self.ask_generator = self._ask(n)
        return next(self.ask_generator)

    def tell(self, config, result):
        self.optimizer_instance.tell(list(config.values()), result)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Importances plot module."""

import numpy as np
import plotly.graph_objs as go

from feijoa import Categorical
from feijoa import create_job
from feijoa import Real
from feijoa import SearchSpace
from feijoa.importance.mdi import MDIEvaluator
from feijoa.importance.rsfanova_boosted import RsFanovaEvaluator


def plot_importances(
    job,
    fig=None,
):
    """Plot the importances for
    a specified job.

    Args:
        job (Job):
            Job instance.
        fig (go.Figure):
            Plotly figure object.

    Raises:
        AnyError: If anything bad happens.

    """

    fig = fig or go.Figure()

    importance_evaluator = MDIEvaluator()
    evaluated = importance_evaluator.do(job)
    params, mdi_importance = (
        evaluated["parameters"],
        evaluated["importance"],
    )
    rsfanova_evaluator = RsFanovaEvaluator()
    rsfanova_importance = rsfanova_evaluator.do(job)["importance"]
    importance = mdi_importance + rsfanova_importance
    importance = importance / 2

    fig.add_bar(
        x=params,
        y=importance,
    )

    fig.update_layout(
        title="Parameters importance plot",
        xaxis_title="Parameter",
        yaxis_title="Relative Importance",
        xaxis={"categoryorder": "total descending"},
    )

    return fig


def objective(experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    z = experiment.params.get("z")
    w = experiment.params.get("w")
    u = experiment.params.get("u")

    ob = x + np.sqrt(y) + np.log(z) + 10 * w
    ob += -10 if u == "bar" else 0
    # ob = x + np.sqrt(y)
    return ob


def main():
    space = SearchSpace()

    space.insert(Real("x", low=0.1, high=3.0))
    space.insert(Real("y", low=0.1, high=3.0))
    space.insert(Real("z", low=0.1, high=3.0))
    space.insert(Real("w", low=0.0, high=3.0))
    space.insert(Categorical("u", choices=["foo", "bar"]))

    job = create_job(search_space=space)
    job.do(
        objective,
        n_jobs=-1,
        n_trials=100,
        algo_list=["bayesian"],
        progress_bar=True,
        use_numba_jit=True,
    )

    fig = plot_importances(job)
    fig.show()


if __name__ == "__main__":
    main()



from feijoa import create_job
from feijoa import Experiment
from feijoa import Real
from feijoa import SearchSpace
from feijoa.models.configuration import Configuration
from feijoa.models.experiment import ExperimentState
from feijoa.storages.rdb.storage import RDBStorage


def test_rdb_storage():
    storage = RDBStorage("sqlite:///:memory:")

    class _Job:
        def __init__(self, job_id, job_name):
            self.id = job_id
            self.name = job_name
            self.search_space = SearchSpace()
            self.search_space.insert(Real("x", low=0.0, high=1.0))
            self.search_space.insert(Real("y", low=0.0, high=1.0))

    ex = Experiment(
        id=0,
        job_id=0,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params=Configuration({"x": 0.0, "y": 1.0}),
    )

    ex_2 = Experiment(
        id=0,
        job_id=1,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params=Configuration({"x": 0.0, "y": 1.0}),
    )

    mock_job = _Job(0, "foo")
    mock_job_2 = _Job(1, "boo")

    storage.insert_job(mock_job)
    storage.insert_job(mock_job_2)

    storage.insert_experiment(ex)
    storage.insert_experiment(ex_2)

    assert storage.get_experiments_by_job_id(0) == [ex]
    assert storage.get_experiments_by_job_id(1) == [ex_2]

    assert storage.get_experiment(job_id=0, experiment_id=0) == ex

    assert storage.jobs == [
        {"id": 0, "name": "foo"},
        {"id": 1, "name": "boo"},
    ]


def test_integrated_rdb_storage():
    space = SearchSpace()
    storage = RDBStorage("sqlite:///:memory:")

    # noinspection DuplicatedCode
    def objective(experiment: Experiment):
        x = experiment.params.get("x")
        y = experiment.params.get("y")
        return (
            (1.5 - x + x * y) ** 2
            + (2.25 - x + x * y**2) ** 2
            + (2.625 - x + x * y**3) ** 2
        )

    space.insert(Real("x", low=0.0, high=5.0))
    space.insert(Real("y", low=0.0, high=2.0))

    job = create_job(search_space=space, storage=storage)
    job.do(objective, n_trials=5)

    print(job.dataframe)



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Job class module."""

import contextlib
from datetime import datetime
from functools import partial
import inspect
import logging
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type
from typing import Union
import warnings

import joblib
import numpy as np
import pandas as pd
from rich.progress import Progress
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError

from feijoa.exceptions import DuplicatedJobError
from feijoa.exceptions import ExperimentNotFinishedError
from feijoa.exceptions import InvalidOptimizer
from feijoa.exceptions import InvalidSearchAlgorithmPassed
from feijoa.exceptions import InvalidStoragePassed
from feijoa.exceptions import InvalidStorageRFC1738
from feijoa.exceptions import JobNotFoundError
from feijoa.exceptions import SearchAlgorithmNotFoundedError
from feijoa.models import Experiment
from feijoa.models import Result
from feijoa.models.experiment import ExperimentState
from feijoa.search import SearchSpace
from feijoa.search.algorithms import get_algo
from feijoa.search.algorithms import SearchAlgorithm
from feijoa.search.algorithms.bayesian import BayesianAlgorithm
from feijoa.search.bandit import ThompsonSampler
from feijoa.search.meta import MetaSearchAlgorithm
from feijoa.search.parameters import Categorical
from feijoa.search.parameters import Integer
from feijoa.search.parameters import Real
from feijoa.search.seed import SeedAlgorithm
from feijoa.storages import Storage
from feijoa.storages.rdb.storage import RDBStorage


__all__ = ["Job", "create_job", "load_job"]

from feijoa.utils.imports import ImportWrapper
from feijoa.utils.misc import in_notebook


log = logging.getLogger(__name__)


class Job:
    """Facade of framework, contains main logic
    of optimization process.

    Example:

        .. code-block:: python

            from feijoa.jobs import Job
            from feijoa.storages.rdb.storage import RDBStorage
            from feijoa.search.space import SearchSpace
            from feijoa.search.bandit import ThompsonSampler

            job = Job(
                name="foo_job",
                storage=RDBStorage("sqlite:///:memory:"),
                search_space=SearchSpace(),
                job_id=1,
                optimizer=ThompsonSampler,
            )

    Args:
        name (str):
            Name of job.
        storage (Storage):
            Storage instance, which will be used to save the results
        search_space (SearchSpace):
            Search space of optimization problem.
        optimizer (MetaSearchAlgorithm):
            Optimizer, which will be used to algorithms manipulation.

    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(
        self,
        name: str,
        storage: Storage,
        search_space: SearchSpace,
        job_id: int,
        pruners: Any = None,
        optimizer=ThompsonSampler,
    ):

        self.name = name
        self.storage = storage
        self.pruners = pruners
        self.id = job_id

        if issubclass(type(optimizer), MetaSearchAlgorithm):
            self.optimizer = optimizer
        elif not issubclass(
            type(optimizer), MetaSearchAlgorithm
        ) and (
            not inspect.isclass(optimizer)
            or not issubclass(optimizer, MetaSearchAlgorithm)
        ):
            raise InvalidOptimizer()
        else:
            self.optimizer = optimizer()

        self.search_space = search_space
        self.pending_experiments = 0

        self.seeds: List[dict] = []

        if not self.storage.is_job_name_exists(self.name):
            self.storage.insert_job(self)

    @property
    def best_parameters(self) -> Optional[dict]:
        """
        Get the best parameters in job.
        Load configurations with result
        and take params from best.

        Returns:
            The best parameters' dict.

        Raises:
            AnyError: If anything bad happens.

        """

        return (
            self.best_experiment.params
            if self.best_experiment
            else None
        )

    @property
    def best_value(self) -> Optional[Any]:
        """
        Get the best result value by objective.

        Returns:
            Float best objective function value.

        Raises:
            AnyError: If anything bad happens.

        """

        return (
            self.best_experiment.objective_result
            if self.best_experiment
            else None
        )

    @property
    def best_experiment(self) -> Optional[Experiment]:
        """
        Get the best experiment from job.

        Returns:
            Best experiment of optimization session.

        Raises:
            AnyError: If anything bad happens.

        """

        return self.storage.best_experiment(self.id)

    @property
    def experiments(self) -> List[Experiment]:
        """
        Get all experiments.

        Returns:
            List of experiments.

        Raises:
            AnyError: If anything bad happens.

        """

        return self.storage.get_experiments_by_job_id(self.id)

    @property
    def experiments_count(self) -> int:
        """
        Get job experiments count.

        Returns:
            Experiments count.

        Raises:
            AnyError: If anything bad happens.
        """

        return self.storage.get_experiments_count(self.id)

    def top_experiments(self, n: int):
        """
        Get top-n experiments by objective.

        Args:
            n (int):
                Max count of top experiments.

        Returns:
            List of top-experiments sorted by objective value.

        Raises:
            AnyError: If anything bad happens.

        """

        return self.storage.top_experiments(self.id, n)

    @property
    def rewards(self):
        """
        Obtaining the results of a strictly
        monotonically decreasing sequence for
        objective function values.

        Returns:
            Length of monotonically decreasing
            sequence for objective function values

        Raises:
            AnyError: If anything bad happens.

        """

        rewards = 0
        experiments = self.experiments
        m = float("+inf")

        for exp in experiments:
            if exp.objective_result < m:
                rewards += 1
                m = exp.objective_result

        return rewards

    def setup_default_algo(self):
        """
        Setup default optimization algorithm (bayesian).

        Raises:
            AnyError: If anything bad happens.

        """

        self.add_algorithm(BayesianAlgorithm(self.search_space))

    def _load_algo(self, algo_list=None):
        """
        Inject algorithms from algo_list to
        current job.

        Args:
            algo_list (List[Union[str, SearchAlgorithm, Type[SearchAlgorithm]]):
                array of algorithms.

        Raises:
            SearchAlgorithmNotFoundedError:
                If passed the name of an algorithm that does not exist.
            InvalidSearchAlgorithmPassed:
                If passes not valid search algorithm.
            AnyError: If anything bad happens.

        """

        if self.seeds:
            self.add_algorithm(SeedAlgorithm(*self.seeds))

        if not algo_list:
            self.add_algorithm(BayesianAlgorithm(self.search_space))
        else:
            for algo in algo_list:
                if isinstance(algo, str):
                    try:
                        algo_cls = get_algo(algo)
                    except SearchAlgorithmNotFoundedError:
                        raise SearchAlgorithmNotFoundedError()
                elif isinstance(algo, SearchAlgorithm):
                    self.add_algorithm(algo)
                    continue
                elif issubclass(algo, SearchAlgorithm):
                    algo_cls = algo
                else:
                    raise InvalidSearchAlgorithmPassed()

                # noinspection PyArgumentList
                self.add_algorithm(
                    algo_cls(search_space=self.search_space)
                )

    def do(
        self,
        objective: Callable,
        n_trials: int = 100,
        n_jobs: int = 1,
        n_points_iter: int = 1,
        algo_list: Optional[
            List[Union[str, Type[SearchAlgorithm]]]
        ] = None,
        clear=True,
        progress_bar=True,
        use_numba_jit=False,
    ):
        """
        Do optimization for current job.

        Example:

            .. code-block:: python

                from feijoa import create_job
                from feijoa import Experiment
                from feijoa import Real
                from feijoa import SearchSpace


                def objective(experiment: Experiment):
                    x = experiment.params.get("x")
                    y = experiment.params.get("y")
                    return (
                        (1.5 - x + x * y) ** 2
                        + (2.25 - x + x * y**2) ** 2
                        + (2.625 - x + x * y**3) ** 2
                    )


                space = SearchSpace()

                space.insert(Real("x", low=0.0, high=5.0))
                space.insert(Real("y", low=0.0, high=2.0))

                job = create_job(search_space=space)
                job.do(objective, n_trials=50)

        Args:
            objective:
                Objective function.
            n_trials (int):
                Number of total runs.
            n_jobs (int):
                Job's count parallelization with `joblib` backend. if -1 passed => used max of CPU's.
            algo_list (List[Union[str, SearchAlgorithm, Type[SearchAlgorithm]]):
                array of algorithms.
            n_points_iter (int):
                The preferred number of configurations in one epoch. May have no effect for some algorithms.
            clear (bool):
                Clear algo list or not.
            progress_bar (bool):
                Show progress bar (rich) or not.
            use_numba_jit (bool):
                Use numba for objective evaluation speedup.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        # noinspection PyPep8Naming

        if clear:
            cls = type(self.optimizer)
            self.optimizer = cls()

        if (
            algo_list
            or algo_list is None
            and not self.optimizer.algorithms
        ):
            self._load_algo(algo_list)

        if use_numba_jit:
            with ImportWrapper():
                from numba import jit

                objective = jit(objective)

        dela = joblib.delayed(objective)

        progress = (
            Progress(transient=True, disable=not progress_bar)
            if progress_bar
            else contextlib.nullcontext()
        )

        trials = 0
        with progress as bar:  # type: ignore
            task = bar.add_task("Optimizing", total=n_trials)
            while trials < n_trials:
                configurations = self.ask(n_points_iter)

                if not configurations:
                    warnings.warn("No new configurations.")
                    # TODO (qnbhd): make closing
                    break

                configurations = configurations[
                    : min(len(configurations), n_trials)
                ]

                if len(configurations) + trials > n_trials:

                    log.warning(
                        f"Requestor: <{configurations[0].params.requestor}>"
                        f" requires a large number of launches to work"
                        f"correctly. this logic will change in the future."
                    )

                    # TODO (qnbhd): fix this logic

                configurations_len = len(configurations)
                trials += configurations_len
                bar.update(task, advance=configurations_len)

                # noinspection PyUnresolvedReferences,PyBroadException
                parallel = joblib.Parallel(
                    n_jobs=n_jobs, prefer="threads"
                )
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    results = parallel(
                        dela(u) for u in configurations
                    )

                # Applying result
                for experiment, result in zip(
                    configurations, results
                ):
                    self.tell(experiment, result)

        if in_notebook():
            from IPython.display import clear_output

            clear_output(wait=False)

    def tell(self, experiment, result: Union[float, Result]):
        """
        Finish concrete experiment
        and results to algorithms.

        Args:
            experiment (Experiment):
                Specified experiment.
            result (Union[float, Result]):
                Result for current experiment.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        objective = (
            result.objective_result
            if isinstance(result, Result)
            else result
        )

        if np.isfinite(objective):
            experiment.success_finish()
        else:
            experiment.error_finish()

        if experiment.is_finished():
            self.pending_experiments -= 1

            if isinstance(result, Result):
                self.optimizer.tell(
                    experiment.params, result.objective_result
                )
                experiment.apply(result.objective_result)
                experiment.metrics = result.metrics
            else:
                self.optimizer.tell(experiment.params, result)
                experiment.apply(result)

            self.storage.insert_experiment(experiment)
            return

        raise ExperimentNotFinishedError()

    def _tell_for_loaded(self, experiment: Experiment):
        """
        Tell results for loaded job.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        self.optimizer.tell(
            experiment.params, experiment.objective_result
        )

    def ask(self, n: int) -> Optional[List[Experiment]]:
        """
        Ask for a new experiment.

        Args:
            n (int):
                Preferred count of experiments:

        .. note::
            n may not affect on experiments count.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        configs = self.optimizer.ask(n)

        if not configs:
            return configs

        applicator = partial(
            Experiment,
            job_id=self.id,
            state=ExperimentState.WIP,
        )

        experiments = [
            applicator(
                params=config,
                id=(
                    self.experiments_count
                    + self.pending_experiments
                    + i
                ),
                create_timestamp=datetime.timestamp(datetime.now()),
            )
            for i, config in enumerate(configs)
        ]

        self.pending_experiments += len(experiments)

        return experiments

    def get_dataframe(self, brief=False, desc=False, only_good=False):
        """Make dataframe for current job.

        Args:
            brief (bool):
                Make brief report without additional information.
            desc (bool):
                Fetch only descending by objective function values
                experiments.
            only_good (bool):
                Fetch only correct (state='OK') experiments.

        Returns:
            Pandas dataframe, contains all information
            about optimization session.

        Raises:
            AnyError: If anything bad happens.

        """

        container = []
        m = float("+inf")

        for experiment in self.experiments:
            dataframe_dict = experiment.dict()

            if dataframe_dict["objective_result"] < m:
                m = dataframe_dict["objective_result"]
            else:
                if desc:
                    continue

            dataframe_dict["state"] = str(dataframe_dict["state"])

            if only_good and dataframe_dict["state"] != "OK":
                continue

            params = dataframe_dict.pop("params")

            metrics = dataframe_dict.pop("metrics") or dict()

            dataframe_dict["create_time"] = datetime.fromtimestamp(
                dataframe_dict["create_timestamp"]
            )

            dataframe_dict["finish_time"] = datetime.fromtimestamp(
                dataframe_dict["finish_timestamp"]
            )

            del dataframe_dict["create_timestamp"]
            del dataframe_dict["finish_timestamp"]
            del dataframe_dict["hash"]
            del dataframe_dict["job_id"]

            if brief:
                idx = dataframe_dict["id"]
                objective_result = dataframe_dict["objective_result"]
                container.append(
                    {
                        "id": idx,
                        "objective_result": objective_result,
                        **params,
                    }
                )
                continue

            container.append({**dataframe_dict, **params, **metrics})

        df = pd.DataFrame(container)

        cols_types = {}
        for col in self.search_space.name2param.keys():
            param = self.search_space.get(col)

            if col not in list(df.columns):
                continue

            if isinstance(param, Categorical):
                cols_types[col] = "category"
            elif isinstance(param, Integer):
                cols_types[col] = "int64"
            elif isinstance(param, Real):
                cols_types[col] = "float32"

        df = df.astype(cols_types)

        return df

    @property
    def dataframe(self):
        """Get dataframe property."""

        return self.get_dataframe()

    def add_algorithm(self, algo: SearchAlgorithm):
        """Add algorithm to job's optimizer

        Arguments:
            algo (SearchAlgorithm):
                search algorithm instance.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        self.optimizer.add_algorithm(algo)

    def add_seed(self, seed):
        """Add seed configuration to current job.

        Args:
            seed (dict):
                Seed configuration,
                which must be measured immediately.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        self.seeds.append(seed)


def _load_storage(
    storage_or_name: Union[str, Optional[Storage]]
) -> Storage:
    """
    Load a storage object.

    Args:
        storage_or_name (Union[str, Optional[Storage]]) :
            Storage instance or string with RFC1738 spec.
    """
    if storage_or_name is None:
        return RDBStorage("sqlite:///:memory:")

    if isinstance(storage_or_name, Storage):
        return storage_or_name

    if not isinstance(storage_or_name, str):
        raise InvalidStoragePassed()

    try:
        url = make_url(storage_or_name)
    except ArgumentError:
        raise InvalidStorageRFC1738()

    assert url.database

    if url.drivername == "tinydb":
        with ImportWrapper():
            from feijoa.storages.tiny import TinyDBStorage

            return TinyDBStorage(str(url.database))

    return RDBStorage(storage_or_name)


def create_job(
    *,
    search_space: SearchSpace,
    name: Optional[str] = None,
    storage: Union[str, Optional[Storage]] = None,
    **kwargs,
):
    """Create job instance
    with specified parameters.

    Example:

        .. code-block:: python

            from feijoa import create_job
            from feijoa import SearchSpace

            space = SearchSpace()

            job = create_job(search_space=space)

    Args:
        search_space (SearchSpace):
            Search space instance.
        name (str):
            Name of job.
        storage:
            Storage instance or string
            with `RFC1738` spec.

    Returns:
        Job instance.

    Raises:
        AnyError: If anything bad happens.

    """

    name = name or "job" + datetime.now().strftime(
        "%H_%M_%S_%m_%d_%Y"
    )

    storage = _load_storage(storage)

    assert isinstance(name, str)
    assert isinstance(search_space, SearchSpace)

    if storage.is_job_name_exists(name):
        raise DuplicatedJobError(f"Job {name} is already exists.")

    return Job(
        name, storage, search_space, storage.jobs_count + 1, **kwargs
    )


def load_job(
    *,
    name: str,
    storage: Optional[Union[str, Storage]] = None,
    **kwargs,
):
    """Load existed job instance
    with specified parameters.

    Example:

        .. code-block:: python

            from feijoa import load_job
            from feijoa import SearchSpace

            # job must be in storage

            job = load_job(name="foo", storage="sqlite:///:memory:")

    Args:
        name (str):
            Name of job.
        storage:
            Storage instance or string
            with `RFC1738` spec.

    Returns:
        Job instance.

    Raises:
        AnyError: If anything bad happens.

    """

    storage = _load_storage(storage)

    job_id = storage.get_job_id_by_name(name)

    if not job_id:
        raise JobNotFoundError(f"Job {name} not found is storage.")

    experiments = storage.get_experiments_by_job_id(job_id)
    search_space = storage.get_search_space_by_job_id(job_id)

    job = Job(name, storage, search_space, job_id, **kwargs)

    for experiment in experiments:
        # noinspection PyProtectedMember
        job._tell_for_loaded(experiment)

    return job



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""fANOVA importances evaluator uses Rust
implementation of fANOVA algorithm.
"""

import numpy as np
from numpy import float64
from sklearn.preprocessing import LabelEncoder

from ..jobs.job import Job
from ..utils.imports import ImportWrapper
from .evaluator import ImportanceEvaluator


with ImportWrapper():
    import rsfanova

__all__ = ["RsFanovaEvaluator"]


class RsFanovaEvaluator(ImportanceEvaluator):
    """fANOVA importance evaluator rust implementation
    binding for https://github.com/sile/fanova

    .. note::
        This evaluator is experimental.

    .. code-block:: python

        from feijoa.importance.rsfanova_boosted import (
            RsFanovaEvaluator,
        )

        job = ...
        evaluator = RsFanovaEvaluator()
        imp = evaluator.do(job)

        params = imp["params"]
        importances = imp["importances"]

    """

    def do(self, job: Job):
        df = job.get_dataframe(brief=True, only_good=True)
        y = df["objective_result"]
        X = df.drop(columns=["objective_result", "id"])
        categorical = X.select_dtypes(include=["category"])
        encoder = LabelEncoder()

        cols = X.columns

        for cat in categorical.columns:
            X[cat] = encoder.fit_transform(X[cat])

        X = np.array(X, dtype=float64)
        y = np.array(y, dtype=float64)
        X = np.ascontiguousarray(X.transpose())

        importance = rsfanova.fetch_importances(X, y)

        completed = dict()
        completed["parameters"] = cols
        completed["importance"] = np.array(importance)
        return completed



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""RDB storage uses SQLAlchemy module."""

from typing import List
from typing import Optional

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from feijoa.models import Experiment
from feijoa.models.configuration import Configuration
from feijoa.search.space import SearchSpace
from feijoa.storages.rdb.models import _Base
from feijoa.storages.rdb.models import ExperimentModel
from feijoa.storages.rdb.models import JobModel
from feijoa.storages.rdb.models import ParameterModel
from feijoa.storages.rdb.models import SearchSpaceModel
from feijoa.storages.storage import Storage


class RDBStorage(Storage):
    """Relational Database Storage.

    Uses SQLAlchemy framework.

    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(self, url):
        self.engine = create_engine(url)
        _Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()

    def insert_job(self, job):
        search_space_model = SearchSpaceModel(id=job.id)
        self.session.add(search_space_model)

        for p in job.search_space:
            name = p.name
            kind = p.__class__.__name__
            meta = p.meta
            param_model = ParameterModel(
                search_space_id=search_space_model.id,
                name=name,
                kind=kind,
                meta=meta,
            )
            self.session.add(param_model)

        job_model = JobModel(
            id=job.id,
            name=job.name,
            search_space_id=search_space_model.id,
        )
        self.session.add(job_model)
        self.session.commit()

    def get_search_space_by_job_id(self, job_id):
        job_model = (
            self.session.query(JobModel).filter_by(id=job_id).one()
        )
        parameters = job_model.search_space.parameters
        return SearchSpace.from_db_parameters(parameters)

    def is_job_name_exists(self, name):
        job_model = (
            self.session.query(JobModel).filter_by(name=name).first()
        )
        return bool(job_model)

    def get_job_id_by_name(self, name) -> Optional[int]:
        job_model = (
            self.session.query(JobModel).filter_by(name=name).first()
        )
        return job_model.id if job_model else None

    def insert_experiment(self, experiment):
        experiment_model = ExperimentModel(
            id=experiment.id,
            job_id=experiment.job_id,
            state=experiment.state,
            hash=experiment.hash,
            objective_result=experiment.objective_result,
            params=experiment.params,
            requestor=experiment.params.requestor,
            create_timestamp=experiment.create_timestamp,
            finish_timestamp=experiment.finish_timestamp,
            metrics=experiment.metrics,
        )
        self.session.add(experiment_model)
        self.session.commit()

    def get_experiment(self, job_id, experiment_id):
        experiment_model = (
            self.session.query(ExperimentModel)
            .filter_by(id=experiment_id, job_id=job_id)
            .one()
        )
        return Experiment.from_orm(experiment_model)

    def get_experiments_by_job_id(self, job_id) -> List[Experiment]:
        experiments_models = (
            self.session.query(ExperimentModel)
            .filter_by(job_id=job_id)
            .all()
        )
        experiments = []
        for exp in experiments_models:
            exp.params = Configuration(
                exp.params, requestor=exp.requestor
            )
            experiments.append(Experiment.from_orm(exp))
        return experiments

    def get_experiments_count(self, job_id) -> int:
        experiments_models = (
            self.session.query(ExperimentModel)
            .filter_by(job_id=job_id)
            .all()
        )
        return len(experiments_models)

    @property
    def jobs(self):
        jobs_models = self.session.query(JobModel).all()
        jobs = []
        for job_m in jobs_models:
            jobs.append({"id": job_m.id, "name": job_m.name})
        return jobs

    @property
    def version(self):
        raise NotImplementedError()

    def __del__(self):
        self.session.close()
        self.engine.dispose()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Round robin meta algorithm module."""

from collections import deque
import logging

from feijoa.search.algorithms.algorithm import SearchAlgorithm
from feijoa.search.meta import MetaSearchAlgorithm


log = logging.getLogger(__name__)

__all__ = ["RoundRobinMeta"]


class RoundRobinMeta(MetaSearchAlgorithm):
    """RoundRobin meta algorithm.

    Raises:
        AnyError: If anything bad happens.

    """

    anchor = "roundrobin"
    aliases = ("roundrobin",)

    def __init__(self, *algorithms):
        super().__init__(*algorithms)
        self.algo_deq = None

    @property
    def order(self):
        # noinspection PyUnresolvedReferences
        if not self.algo_deq:
            self.algo_deq = deque(self.algorithms)

        self.algo_deq.rotate(1)
        return self.algo_deq

    def remove_algorithm(self, algo: SearchAlgorithm):
        self.algo_deq.remove(algo)
        self.algorithms.remove(algo)



import pytest

from feijoa import Categorical
from feijoa import Integer
from feijoa import Real
from feijoa.exceptions import ParametersIncorrectInputValues
from feijoa.search.parameters import ParametersVisitor


def test_parameters_constructing():

    with pytest.raises(ParametersIncorrectInputValues):
        Integer("x", low="foo", high="bar")

    with pytest.raises(ParametersIncorrectInputValues):
        Integer("x", low=0.1, high=10.0)

    with pytest.raises(ParametersIncorrectInputValues):
        Real("x", low="foo", high="bar")

    with pytest.raises(ParametersIncorrectInputValues):
        Real("x", low=1, high=2)

    with pytest.raises(ParametersIncorrectInputValues):
        Categorical("x", choices=[])

    with pytest.raises(ParametersIncorrectInputValues):
        Categorical("x", choices=["foo", "foo"])

    with pytest.raises(ParametersIncorrectInputValues):
        Integer("x", low=10, high=1)

    with pytest.raises(ParametersIncorrectInputValues):
        Real("x", low=10.0, high=1.0)


def test_parameters_seed():

    i = Integer("x", low=0, high=10)
    assert i.seed() == 0

    r = Real("y", low=1.0, high=10.0)
    assert r.seed() == 1.0

    c = Categorical("z", choices=["foo", "bar"])
    assert c.seed() == "foo"


def test_parameters_visitor():
    class FooVisitor(ParametersVisitor):
        def visit_integer(self, p: Integer):
            return f"integer {p.name}"

        def visit_real(self, p: Real):
            return f"real {p.name}"

        def visit_categorical(self, p: Categorical):
            return f"categorical {p.name}"

    parameters = [
        Integer("x", low=0, high=10),
        Real("y", low=0.0, high=1.0),
        Categorical("z", choices=["foo", "bar"]),
    ]

    visitor = FooVisitor()
    ans = [p.accept(visitor) for p in parameters]

    assert ans == ["integer x", "real y", "categorical z"]



from feijoa import Categorical
from feijoa import create_job
from feijoa import Experiment
from feijoa import Integer
from feijoa import Real
from feijoa import SearchSpace


def objective(experiment: Experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    z = experiment.params.get("z")

    a = 0
    if z == "foo":
        a = 1
    if z == "bar":
        a = -1

    return (
        (1.5 - x + x * y) ** 2
        + (2.25 - x + x * y**2) ** 2
        + (2.625 - x + x * y**3) ** 2
        + a
    )


def test_template_search():
    space = SearchSpace()

    space.insert(Real("x", low=0.0, high=5.0))
    space.insert(Integer("y", low=0, high=2))
    space.insert(Categorical("z", choices=["foo", "bar"]))

    job = create_job(search_space=space)
    job.do(objective, n_trials=200, algo_list=["template"])



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""TinyDB storage module."""

from typing import List
from typing import NamedTuple
from typing import Optional

from feijoa.utils.imports import ImportWrapper


with ImportWrapper():
    from tinydb import Query, TinyDB
    from tinydb.table import Document

from feijoa.exceptions import DBVersionError
from feijoa.exceptions import InsertExperimentWithTheExistedId
from feijoa.models import Experiment
from feijoa.models.configuration import Configuration
from feijoa.search.space import SearchSpace
from feijoa.storages.storage import Storage


__all__ = ["TinyDBStorage"]


class _Parameter(NamedTuple):
    name: str
    kind: str
    meta: dict


# noinspection PyTypeChecker
class TinyDBStorage(Storage):
    """TinyDB storage uses tinydb library."""

    __version__ = "0.1.0"

    def __init__(self, json_file: str):
        self.json_file = json_file
        self.tiny_db = TinyDB(json_file)

        db_version_table = self.tiny_db.table("version")

        if not len(db_version_table.all()):
            # New table
            db_version_table.insert({"version": self.version})
        else:
            ver_record, *_ = db_version_table.all()
            ver = ver_record["version"]
            if ver < self.version:
                raise DBVersionError()

        self.jobs_table = self.tiny_db.table("job")
        self.experiments_table = self.tiny_db.table("experiment")
        self.parameters_table = self.tiny_db.table("parameters")

    def insert_job(self, job):
        doc = {
            "name": job.name,
            "id": job.id,
        }

        self.jobs_table.insert(doc)

        for p in job.search_space:
            name = p.name
            kind = p.__class__.__name__
            meta = p.meta
            param_model = {
                "job_id": job.id,
                "name": name,
                "kind": kind,
                "meta": meta,
            }
            self.parameters_table.insert(param_model)

    def get_search_space_by_job_id(self, job_id) -> SearchSpace:
        parameters = self.parameters_table.search(
            Query().job_id == job_id
        )
        pool = list()
        for p in parameters:
            mod = _Parameter(
                name=p["name"], kind=p["kind"], meta=p["meta"]
            )
            pool.append(mod)
        return SearchSpace.from_db_parameters(pool)

    def get_job_id_by_name(self, name) -> Optional[int]:
        jobs = self.jobs_table.search(Query().name == name)

        if not len(jobs):
            return None

        assert len(jobs) == 1

        job, *_ = jobs

        return job["id"]

    def is_job_name_exists(self, name):
        jobs = self.jobs_table.search(Query().name == name)
        return len(jobs) != 0

    def insert_experiment(self, experiment):
        doc = experiment.dict()
        doc["requestor"] = experiment.params.requestor

        if self.get_experiment(experiment.job_id, experiment.id):
            # TODO (qnbhd): make correct exception
            raise InsertExperimentWithTheExistedId()

        self.experiments_table.insert(doc)

    def get_experiment(self, job_id, experiment_id):
        q = self.experiments_table.search(
            (Query().id == experiment_id) & (Query().job_id == job_id)
        )

        if not q:
            return None

        exp = q[0]
        exp["params"] = Configuration(
            exp["params"], requestor=exp["requestor"]
        )
        exp.pop("requestor")

        return Experiment(**exp)

    def _get_raw_experiments(self, job_id) -> List[Document]:
        docs = self.experiments_table.search(Query().job_id == job_id)
        return docs

    def get_experiments_by_job_id(self, job_id) -> List[Experiment]:
        docs = self._get_raw_experiments(job_id)
        experiments = []
        for doc in docs:
            doc["params"] = Configuration(
                doc["params"], requestor=doc["requestor"]
            )
            # doc.pop('requestor')
            exp = Experiment(**doc)
            experiments.append(exp)
        return experiments

    def get_experiments_count(self, job) -> int:
        return len(self.get_experiments_by_job_id(job))

    @property
    def jobs(self):
        return self.jobs_table.all()

    @property
    def version(self):
        return self.__version__

    def __del__(self):
        self.tiny_db.close()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Search space module."""

from typing import List
import warnings

import yaml

from feijoa.search.parameters import Categorical
from feijoa.search.parameters import Integer
from feijoa.search.parameters import Parameter
from feijoa.search.parameters import Real


__all__ = ["SearchSpace"]


class SearchSpace:
    """Search space class.

    Represents cartesian multiplication of
    set of available values for every
    parameter.

    Example:

        .. code-block:: python

            from feijoa.search.space import SearchSpace
            from feijoa.search.parameters import Real

            space = SearchSpace()
            space.insert(Real("x", low=0.0, high=1.0))

    Args:
        parameters (list, optional):
            List of parameters instances.


    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(self, *parameters):
        self.params: List[Parameter] = parameters or list()
        self.name2param = {p.name: p for p in parameters} or dict()

    def insert(self, p: Parameter) -> None:
        """Insert parameter into search space"""

        # TODO (qnbhd): make duplication check
        self.params.append(p)
        self.name2param[p.name] = p

    def get(self, item: str, default=None):
        """Get parameters by name from search space."""

        return self.name2param.get(item, default)

    def __repr__(self) -> str:
        buff = ["SearchSpace:"]

        for p in self.params:
            buff.append(f"{' ' * 4}{p}")

        return "\n".join(buff)

    def __str__(self):
        return repr(self)

    def __iter__(self):
        return iter(self.params)

    def __len__(self):
        return len(self.params)

    @classmethod
    def from_yaml(cls, yaml_string):
        """Load search space from yaml."""

        space_dumped = yaml.safe_load(yaml_string)

        space = SearchSpace()

        for p in space_dumped:
            loaded = parameter_factory(**p)
            space.insert(loaded)

        return space

    @classmethod
    def from_yaml_file(cls, yaml_file):
        """Load search space from yaml file."""

        with open(yaml_file) as f:
            return cls.from_yaml(f.read())

    @classmethod
    def from_db_parameters(cls, parameters):
        """Build search space from db representation
        of parameters.

        Raises:
            AnyError: If anything bad happens.

        """

        pool = list()

        for param in parameters:
            name = param.name
            kind = param.kind
            meta = param.meta
            kwargs = {
                "signature": name,
                "type": kind,
                **meta,
            }
            p = parameter_factory(**kwargs)
            pool.append(p)

        return SearchSpace(*pool)


def parameter_factory(**kwargs):
    """Create parameters from dict representation.

    Args:
        kwargs (Dict[str]):
            Dict with information about parameter.

    Returns:
        Parameter builded from kwargs.

    Raises:
        AnyError: If anything bad happens.

    """

    param_signature = kwargs["signature"]

    if kwargs["type"] in ("integer", "Integer"):
        param_potentially_range = kwargs.get("range", None)
        if param_potentially_range is not None:
            low, high = param_potentially_range
        else:
            low, high = kwargs["low"], kwargs["high"]
        return Integer(param_signature, low=low, high=high)
    elif kwargs["type"] in ("real", "Real"):
        param_potentially_range = kwargs.get("range", None)
        if param_potentially_range is not None:
            low, high = param_potentially_range
        else:
            low, high = kwargs["low"], kwargs["high"]
        return Real(param_signature, low=low, high=high)
    elif kwargs["type"] in ("categorical", "Categorical"):
        return Categorical(param_signature, choices=kwargs["choices"])

    raise TypeError()


def from_yaml(yaml_file: str) -> SearchSpace:
    """Load search space from yaml file."""

    warnings.warn(
        "Function `from_yaml` is deprecated."
        "Use SearchSpace.from_yaml instead.",
        DeprecationWarning,
    )

    with open(yaml_file) as f:
        space_dumped = yaml.safe_load(f)

    space = SearchSpace()

    for p in space_dumped:
        loaded = parameter_factory(**p)
        space.insert(loaded)

    return space



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from datetime import datetime
import json
import logging
import sys

import click

from feijoa.utils.logging import init
from utils.extractor import extract
from utils.run_tools import continue_job
from utils.run_tools import run_job


print(sys.path)

now = datetime.now().strftime("%H-%M-%S_%m_%d_%Y")

init(verbose=True)

log = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@click.command(name="run")
@click.option("--toolchain", type=str, required=True)
@click.option("--search-space", type=str, required=True)
@click.option("--source-file", type=str, required=True)
@click.option("--iterations", type=int, default=5)
@click.option("--n-trials", type=int, default=100)
@click.option("--storage", type=str, default=f"sqlite:///{now}.db")
@click.option("--job-name", type=str, default=now)
def run_cmd(
    toolchain,
    search_space,
    source_file,
    n_trials,
    iterations,
    storage,
    job_name,
):
    init(verbose=True)
    baselines, job = run_job(
        toolchain,
        search_space,
        source_file,
        n_trials,
        iterations,
        storage,
        job_name,
        "time",
    )
    log.info("Baselines:")
    log.info(json.dumps(baselines, indent=2))
    print(job.dataframe)


@click.command(name="continue")
@click.option("--toolchain", type=str, required=True)
@click.option("--source-file", type=str, required=True)
@click.option("--iterations", type=int, default=5)
@click.option("--storage", type=str, required=True)
@click.option("--job-name", type=str, required=True)
def continue_cmd(
    toolchain,
    source_file,
    n_trials,
    iterations,
    storage,
    job_name,
):
    init(verbose=True)
    baselines, job = continue_job(
        toolchain,
        source_file,
        n_trials,
        iterations,
        storage,
        job_name,
        "time",
    )
    log.info("Baselines:")
    log.info(baselines)
    print(job.dataframe)


@click.command(name="extract")
@click.option("--toolchain", type=str, required=True)
def extract_cmd(toolchain):
    extract(toolchain)


cli.add_command(run_cmd)
cli.add_command(continue_cmd)
cli.add_command(extract_cmd)

if __name__ == "__main__":
    cli()



from feijoa import Categorical
from feijoa import create_job
from feijoa import Experiment
from feijoa import Integer
from feijoa import Real
from feijoa import SearchSpace


def faked_random(nums):
    f = fake_random(nums)

    def inner(*args, **kwargs):
        return next(f)

    return inner


def fake_random(nums):
    i = 0
    while True:
        yield nums[i]
        i = (i + 1) % len(nums)


def test_random_search():
    space = SearchSpace()
    space.insert(Real("x", low=0.0, high=1.0))
    space.insert(Real("y", low=0.0, high=1.0))
    space.insert(Integer("z", low=0, high=2))
    space.insert(Categorical("w", choices=["foo"]))

    def objective(experiment: Experiment):
        params = experiment.params

        x = params.get("x")
        y = params.get("y")

        return (1 - x) ** 2 + (1 - y) ** 2

    job = create_job(search_space=space)
    job.do(objective, n_trials=10, algo_list=["random"])

    assert job.best_parameters == {
        "w": "foo",
        "x": 0.8444218515250481,
        "y": 0.7579544029403025,
        "z": 1,
    }



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [
    "DBVersionError",
    "SearchAlgorithmNotFoundedError",
    "DuplicatedJobError",
    "JobNotFoundError",
    "ExperimentNotFinishedError",
    "ParametersIncorrectInputValues",
    "InsertExperimentWithTheExistedId",
    "InvalidStoragePassed",
    "InvalidStorageRFC1738",
    "PackageNotInstalledError",
    "InvalidOptimizer",
]


class FeijoaError(Exception):
    """Common class for feijoa exceptions."""


class DBVersionError(FeijoaError):
    """
    Raises if the loaded storage version
    does not match the current version
    """


class SearchAlgorithmNotFoundedError(FeijoaError):
    """
    Raises if the chosen search algorithm
    not founded.
    """


class InvalidSearchAlgorithmPassed(FeijoaError):
    """Raises if the specified object is
    not valid SearchAlgorithm.
    """


class DuplicatedJobError(FeijoaError):
    """
    Raises if the specified job name already
    exists in the storage.
    """


class JobNotFoundError(FeijoaError):
    """
    Raises if the specified job name not
    exists in the storage.
    """


class ExperimentNotFinishedError(FeijoaError):
    """
    Raises if an attempt was made to inform
    the search algorithms that the experiment
    was not completed. To complete the experiment,
    you must call the .success_finish() method
    """


class ParametersIncorrectInputValues(FeijoaError):
    """
    Raises if incorrect values passed to parameter's constructor.
    """


class InsertExperimentWithTheExistedId(FeijoaError):
    """
    Raises if the specified experiment with current id
    is already exists in db storage.
    """


class InvalidStoragePassed(FeijoaError):
    """Raises if the specified object is not a storage."""


class InvalidStorageRFC1738(FeijoaError):
    """Raises if you could not parse rfc1738 URL
    from specified string."""


class InvalidOptimizer(FeijoaError):
    """Raises if specified optimizer class doesn't
    inherit from MetaSearchAlgorithm"""


class PackageNotInstalledError(FeijoaError):
    """Raises if an attempt to import a package
    was not successful"""



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import json
import logging
import os
from os.path import abspath
from os.path import dirname
import re

from feijoa.utils.imports import ImportWrapper


with ImportWrapper():
    import executor

log = logging.getLogger(__name__)

APPLICATION = os.path.join(
    dirname(dirname(abspath(__file__))), "raytracer", "raytracer.cpp"
)


def check_is_working_flag(gcc_toolchain_path, flag):
    """Check to see if a given flag is currently running.

    Args:
        gcc_toolchain_path (str):
            GCC compiler path.
        flag (str):
            Flag to test.

    Raises:
        AnyError: If anything bad happens.

    """

    cmd = f"{gcc_toolchain_path} {APPLICATION} {flag} 2>&1"
    try:
        result = executor.execute(cmd, capture=True)
    except executor.ExternalCommandFailed:
        return False
    return result == ""


def invert_gcc_flag(flag_):
    """Invert a gcc flag

    Args:
        flag_ (str):
            Flag to invert.

    Raises:
        AnyError: If anything bad happens.

    """

    assert flag_[:2] == "-f"
    if flag_[2:5] != "no-":
        return "-fno-" + flag_[2:]
    return "-f" + flag_[5:]


def parse_optimizers(gcc_toolchain_path):
    """Parse GCC optimizers.

    Args:
        gcc_toolchain_path (str):
            GCC compiler path.

    Raises:
        AnyError: If anything bad happens.

    """

    try:
        result = executor.execute(
            f"{gcc_toolchain_path} --help=optimizers -Q", capture=True
        )
    except executor.ExternalCommandFailed:
        return {}

    out = result.split("\n")[1:]

    O_num_pat = re.compile("-O<number>")
    flag_align_eq_pat = re.compile("-f(align-[-a-z]+)=")
    flag_pat = re.compile("-f([-a-z0-9]+)")
    flag_enum_pat = re.compile(
        "-f([-a-z0-9]+)=\\[([-A-Za-z_\\|]+)\\]"
    )
    flag_interval_pat = re.compile(
        "-f([-a-z0-9]+)=<([0-9]+),([0-9]+)>"
    )
    flag_number_pat = re.compile("-f([-a-z0-9]+)=<number>")

    optimizers = {}

    def parse_line(line):
        bits = line.split()
        if not bits:
            return

        spec = bits[0]

        m = O_num_pat.fullmatch(spec)
        if m:
            choices = ["-O0", "-O1", "-O2", "-O3"]
            optimizers["-O"] = {
                "type": "categorical",
                "choices": choices,
            }
            return

        # -falign-str=
        m = flag_align_eq_pat.fullmatch(spec)
        if m:
            choices = []
            name = "-f" + m.group(1)

            if check_is_working_flag(gcc_toolchain_path, name):
                choices.append(name)

            inverted = invert_gcc_flag(name)

            if check_is_working_flag(gcc_toolchain_path, inverted):
                choices.append(inverted)

            choices.append(None)

            optimizers[name] = {
                "type": "categorical",
                "choices": choices,
            }
            return

        # -fflag
        m = flag_pat.fullmatch(spec)
        if m:
            choices = []
            name = "-f" + m.group(1)

            if check_is_working_flag(gcc_toolchain_path, name):
                choices.append(name)

            inverted = invert_gcc_flag(name)

            if check_is_working_flag(gcc_toolchain_path, inverted):
                choices.append(inverted)

            choices.append(None)

            optimizers[name] = {
                "type": "categorical",
                "choices": choices,
            }
            return

        # -fflag=[a|b]
        m = flag_enum_pat.fullmatch(spec)
        if m:
            name = "-f" + m.group(1)
            values = m.group(2).split("|")
            values = [f"{name}={v}" for v in values]
            optimizers[name] = {
                "type": "categorical",
                "choices": values,
            }
            return

        # -fflag=<min, max>
        m = flag_interval_pat.fullmatch(spec)
        if m:
            name = "-f" + m.group(1)
            minimum = int(m.group(2))
            maximum = int(m.group(3))
            if minimum < maximum:
                optimizers[name] = {
                    "type": "integer",
                    "range": [minimum, maximum],
                }
            return

        # -fflag=<number>
        m = flag_number_pat.fullmatch(spec)
        if m:
            name = "-f" + m.group(1)
            minimum = 0
            maximum = 2 << 31 - 1
            optimizers[name] = {
                "type": "integer",
                "range": [minimum, maximum],
            }
            return

        log.warning(f"Unknown optimizer {line}")

    for line_ in out:
        parse_line(line_)

    return optimizers


def parse_parameters(gcc_toolchain_path):
    """Parse the command line GCC parameters.

    Args:
        gcc_toolchain_path (str):
            GCC compiler path.

    Raises:
        AnyError: If anything bad happens.

    """

    try:
        result = executor.execute(
            f"{gcc_toolchain_path} --help=params -Q", capture=True
        )
    except executor.ExternalCommandFailed:
        return {}

    out = result.split("\n")[1:]

    param_enum_pat = re.compile(
        "--param=([-a-zA-Z0-9]+)=\\[([-A-Za-z_\\|]+)\\]"
    )
    param_interval_pat = re.compile(
        "--param=([-a-zA-Z0-9]+)=<(-?[0-9]+),([0-9]+)>"
    )
    param_number_pat = re.compile("--param=([-a-zA-Z0-9]+)=")
    param_old_interval_pat = re.compile(
        "\\s+([-a-zA-Z0-9]+)\\s+default\\s+(-?\\d+)\\s"
        "+minimum\\s+(-?\\d+)\\s+maximum\\s+(-?\\d+)"
    )

    params = {}

    def parse_line(line: str):
        """Parse a single line"""

        bits = line.split()
        if not bits:
            return

        if len(bits) <= 1:
            return

        if len(bits) > 2:
            spec = line
        else:
            spec = bits[0]

        spec = spec.strip()

        # --param=name=[a|b]
        m = param_enum_pat.fullmatch(spec)
        if m:
            name = "--param=" + m.group(1)
            values = m.group(2).split("|")
            values = [f"{name}={v}" for v in values]

            params[name] = {"type": "categorical", "choices": values}
            log.info(f"Categorical: {name} {values}")
            return

        m = param_interval_pat.fullmatch(spec)
        if m:
            name = "--param=" + m.group(1)
            minimum = int(m.group(2))
            maximum = int(m.group(3))
            if minimum < maximum:
                params[name] = {
                    "type": "integer",
                    "range": [minimum, maximum],
                }
            log.info(f"Integer: {name} {minimum} {maximum}")
            return

        # --param=name=
        m = param_number_pat.fullmatch(spec)
        if m:
            name = "--param=" + m.group(1)
            minimum = 0
            maximum = 2 << 31 - 1

            params[name] = {
                "type": "integer",
                "range": [minimum, maximum],
            }
            log.info(f"Integer: {name} {minimum} {maximum}")
            return

        m = param_old_interval_pat.fullmatch(spec)
        if m:
            name = "--param=" + m.group(1)
            minimum = int(m.group(3))
            maximum = int(m.group(4))
            if minimum < maximum:
                params[name] = {
                    "type": "integer",
                    "range": [minimum, maximum],
                }
                log.info(f"Integer: {name} {minimum} {maximum}")
                return
            else:
                log.info(
                    f"Parameter {m.group(1)} has incorrect bounds."
                )

        log.warning(f"Unknown parameter {line}")

    for line_ in out:
        parse_line(line_)

    return params


cc_black_list_chunks = [
    "fipa",
    "live-patching",
    "logical-op-non-short-circuit",
    "prefetch-minimum-stride",
    "sched-autopref-queue-depth",
    "vect-max-peeling-for-alignment",
    "handle-exceptions",
    "fsave-optimization-record",
    "flive",
    "fassociative-math",
    "stack-protector",
    "threadsafe-statics",
    "pack-struct",
    "fstack-check",
    "clash_protection",
    "asan",
]


def fix_parameters(gcc_toolchain, all_params):
    """This function fixes the parameters in the GCC

    Args:
        gcc_toolchain (str):
            GCC compiler path.
        all_params (dict):
            All parameters space.

    Raises:
        AnyError: If anything bad happens.

    """

    def keep(param, value):
        """Helper function."""

        for bcc in cc_black_list_chunks:
            if bcc in param:
                return False

        if value["type"] == "integer":
            if not check_is_working_flag(
                gcc_toolchain, f'{param}={value["range"][0]}'
            ):
                return False
        if value["type"] == "categorical":
            if not check_is_working_flag(
                gcc_toolchain, value["choices"][0]
            ):
                return False

        return True

    fixed = {p: v for p, v in all_params.items() if keep(p, v)}

    return fixed


def get_version(gcc_toolchain_path):
    """Get the GCC version.

    Args:
        gcc_toolchain_path (str):
            GCC compiler path.

    Raises:
        AnyError: If anything bad happens.

    """
    try:
        result = executor.execute(
            f"{gcc_toolchain_path} --version", capture=True
        )
        spl = (
            result.split("\n")[0]
            .replace(" ", "_")
            .replace("(", "_")
            .replace(")", "_")
            .replace(".", "_")
        )
        spl = spl.replace("__", "_")
        return spl
    except executor.ExternalCommandFailed:
        return ""


def extract(gcc_toolchain_path):
    """Extract optimization's flags."""

    version = get_version(gcc_toolchain_path)
    if not version:
        raise RuntimeError("Toolchain isn't runnable")

    if not os.path.exists(f"{version}_params.json"):
        params = parse_parameters(gcc_toolchain_path)

        with open(f"{version}_params.json", "w") as f:
            f.write(json.dumps(params, indent=2))
    else:
        with open(f"{version}_params.json") as f:
            params = json.loads(f.read())

    if not os.path.exists(f"{version}_optimizers.json"):
        optimizers = parse_optimizers(gcc_toolchain_path)

        with open(f"{version}_optimizers.json", "w") as f:
            f.write(json.dumps(optimizers, indent=2))
    else:
        with open(f"{version}_optimizers.json") as f:
            optimizers = json.loads(f.read())

    all_optimizations = {**params, **optimizers}

    fixed = fix_parameters(gcc_toolchain_path, all_optimizations)

    print(fixed)

    template = """- signature: {}
  type: {}
  choices: {}\n\n"""

    integer_template = """- signature: {}
  type: {}
  range: {}\n\n"""

    out = open(f"{version}_search_space.yaml", "w")

    for k, v in fixed.items():
        flag_type = v["type"]
        if flag_type == "categorical":
            choices = (
                "["
                + ", ".join(
                    [u if u else "null" for u in v["choices"]]
                )
                + "]"
            )
            out.write(template.format(k, flag_type, choices))
        elif flag_type == "integer":
            out.write(
                integer_template.format(k, flag_type, v["range"])
            )

    out.close()



import unittest.mock

import pytest

from feijoa import Experiment
from feijoa import Real
from feijoa import SearchSpace

# noinspection DuplicatedCode
from feijoa.exceptions import DBVersionError
from feijoa.exceptions import InsertExperimentWithTheExistedId
from feijoa.models.configuration import Configuration
from feijoa.models.experiment import ExperimentState
from feijoa.utils.imports import import_or_skip


TinyDBStorage = import_or_skip("feijoa.storages.tiny").TinyDBStorage


def test_tiny_db_version_check():
    TinyDBStorage("boo.json")

    with unittest.mock.patch.object(
        TinyDBStorage, "__version__", "0.1.1"
    ):
        with pytest.raises(DBVersionError):
            TinyDBStorage("boo.json")


def test_insert_experiment_with_the_same_id():
    storage = TinyDBStorage("woo.json")

    ex = Experiment(
        id=0,
        job_id=0,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params=Configuration({"x": 0.0, "y": 1.0}),
    )

    ex_duplicated = Experiment(
        id=0,
        job_id=0,
        state=ExperimentState.WIP,
        create_timestamp=0.0,
        params=Configuration({"x": 0.0, "y": 1.0}),
    )

    storage.insert_experiment(ex)

    with pytest.raises(InsertExperimentWithTheExistedId):
        storage.insert_experiment(ex_duplicated)

    assert storage.get_experiment(0, 1000) is None


def test_load_job_by_id():
    storage = TinyDBStorage("simple.json")

    class _Job:
        def __init__(self, job_id, job_name):
            self.id = job_id
            self.name = job_name
            self.search_space = SearchSpace()
            self.search_space.insert(Real("x", low=0.0, high=1.0))
            self.search_space.insert(Real("y", low=0.0, high=1.0))

    job = _Job(0, "foo")

    storage.insert_job(job)

    assert storage.get_job_id_by_name("foo") == 0



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Histogram's plot module."""

import plotly.graph_objs as go

from feijoa import create_job
from feijoa import Real
from feijoa import SearchSpace


def plot_objective_hist(
    job,
    fig=None,
    invert_objective=False,
):
    """Plot objective hists for
    a specified job.

    Args:
        job (Job):
            Job instance.
        fig (go.Figure):
            Plotly figure object.
        invert_objective (bool):
            For the maximization task, it is
            necessary to invert the value of
            the view function, in this case,
            you must check this box to get
            corrective values

    Raises:
        AnyError: If anything bad happens.

    """

    fig = fig or go.Figure()
    df = job.get_dataframe(brief=True)
    objectives = df["objective_result"] * (
        1 if not invert_objective else -1
    )

    fig.add_histogram(x=objectives, nbinsx=100)

    fig.update_layout(
        title="Objective result histogram",
        xaxis_title="Objective value",
        yaxis_title="Count",
    )

    return fig


def objective(experiment):
    x = experiment.params.get("x")
    y = experiment.params.get("y")
    return (
        (1.5 - x + x * y) ** 2
        + (2.25 - x + x * y**2) ** 2
        + (2.625 - x + x * y**3) ** 2
    )


def main():
    space = SearchSpace()

    space.insert(Real("x", low=0.0, high=3.0))
    space.insert(Real("y", low=0.0, high=1.0))

    job = create_job(search_space=space)
    job.do(
        objective,
        n_jobs=-1,
        n_trials=2000,
        algo_list=["grid"],
        progress_bar=True,
        use_numba_jit=True,
    )

    fig = plot_objective_hist(job)
    fig.show()


if __name__ == "__main__":
    main()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Configuration model class module."""

from pprint import pformat


__all__ = ["Configuration"]


class Configuration(dict):
    """Configuration model, inherited from built-in dict.

    Example:
        .. code-block:: python

            from feijoa.model.configuration import Configuration

            configuration = Configuration(
                {"foo": 1, "bar": 2}, requestor="simple", request_id=0
            )

    Args:
        1st:
            configuration dict.
        requestor (str, optional):
            Name of requestor (algorithm).
        request_id (int, optional):
            Index of requested configuration
            for specified requestor.

    """

    def __init__(
        self,
        *args,
        requestor="UNKNOWN",
        request_id=0,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.requestor = requestor
        self.request_id = request_id

    def __str__(self):
        fmt = pformat(dict(self.items()))
        return (
            f"Configuration({fmt}, "
            f"requestor={self.requestor}, "
            f"request_id={self.request_id})"
        )

    def __repr__(self):
        return self.__str__()



from typing import List
from typing import Optional

import pytest

from feijoa import create_job
from feijoa import Experiment
from feijoa import load_job
from feijoa import Real
from feijoa import SearchSpace
from feijoa.exceptions import DuplicatedJobError
from feijoa.exceptions import InvalidSearchAlgorithmPassed
from feijoa.exceptions import InvalidStoragePassed
from feijoa.exceptions import InvalidStorageRFC1738
from feijoa.exceptions import JobNotFoundError
from feijoa.exceptions import SearchAlgorithmNotFoundedError
from feijoa.jobs.job import _load_storage
from feijoa.models import Result
from feijoa.search.algorithms import SearchAlgorithm


def test_job():
    job = create_job(search_space=SearchSpace())

    job.setup_default_algo()

    assert job.best_experiment is None
    assert job.best_value is None
    assert job.best_parameters is None
    assert isinstance(job.experiments, list) and not job.experiments
    assert job.experiments_count == 0
    assert job.top_experiments(10) == []
    assert job.dataframe.empty


def test_create_load_job():
    space = SearchSpace()
    space.insert(Real("x", low=0.0, high=1.0))
    space.insert(Real("y", low=0.0, high=1.0))

    def objective(experiment: Experiment):
        params = experiment.params

        x = params.get("x")
        y = params.get("y")

        metrics = {
            "foo": 1.0,
            "bar": 2.0,
        }

        obj = (1 - x) ** 2 + (1 - y) ** 2

        return Result(objective_result=obj, metrics=metrics)

    job = create_job(
        search_space=space, name="foo", storage="sqlite:///foo.db"
    )
    job.do(objective, n_trials=10, algo_list=["random"])

    assert (
        isinstance(job.experiments, list)
        and len(job.experiments) == 10
    )
    assert job.experiments_count == 10
    assert len(job.top_experiments(100)) == 10

    print(job.dataframe)

    job2 = load_job(name="foo", storage="sqlite:///foo.db")

    assert (
        isinstance(job2.experiments, list)
        and len(job2.experiments) == 10
    )
    assert job2.experiments_count == 10
    assert len(job2.top_experiments(100)) == 10

    print(job2.get_dataframe(desc=True))


def test_incorrect_algo_passed():
    space = SearchSpace()
    job = create_job(search_space=space)

    with pytest.raises(SearchAlgorithmNotFoundedError):
        job.do(lambda: 1, algo_list=["some_incorrect"])

    class IncorrectAlgo:
        def ask(self) -> Optional[List[Experiment]]:
            pass

        def tell(self, experiment: Experiment):
            pass

    with pytest.raises(InvalidSearchAlgorithmPassed):
        job.do(lambda: 1, algo_list=[IncorrectAlgo])


def test_correct_algo_subclass_passed():
    space = SearchSpace()
    job = create_job(search_space=space)

    class CorrectAlgo(SearchAlgorithm):

        anchor = "CorrectAlgo"
        aliases = ("CorrectAlgo",)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def ask(self, n: int = 1) -> Optional[List[Experiment]]:
            pass

        def tell(self, config, result):
            pass

    job.do(lambda: 1, algo_list=[CorrectAlgo])


def test_no_configurations_warning():
    space = SearchSpace()
    job = create_job(search_space=space)

    class SomeAlgo(SearchAlgorithm):

        anchor = "SomeAlgo"
        aliases = ("SomeAlgo",)

        def ask(self, n: int = 1) -> Optional[List[Experiment]]:
            return None

        def tell(self, config, result):
            pass

    with pytest.warns(UserWarning, match="No new configurations."):
        job.do(space, algo_list=[SomeAlgo()])


def test_job_duplicated_error():
    create_job(
        search_space=SearchSpace(),
        name="foo",
        storage="sqlite:///foo.db",
    )

    with pytest.raises(DuplicatedJobError):
        create_job(
            search_space=SearchSpace(),
            name="foo",
            storage="sqlite:///foo.db",
        )


def test_load_non_existed():

    with pytest.raises(JobNotFoundError):
        load_job(name="dimple", storage="sqlite:///:memory:")


def test_load_storage():
    with pytest.raises(InvalidStorageRFC1738):
        _load_storage("dfddsfdsfsd")

    with pytest.raises(InvalidStoragePassed):
        _load_storage(dict(a=1))



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import contextlib
from functools import lru_cache
import importlib.util
import inspect
from itertools import chain
import pathlib
import sys

from feijoa import __feijoa_folder__
from feijoa.exceptions import PackageNotInstalledError
from feijoa.exceptions import SearchAlgorithmNotFoundedError
from feijoa.search.algorithms.algorithm import SearchAlgorithm


ALGORITHMS_FOLDER = (
    pathlib.Path(__feijoa_folder__) / "search" / "algorithms"
)
INTEGRATION_FOLDER = (
    pathlib.Path(__feijoa_folder__).parent
    / "integration"
    / "algorithms"
)


@lru_cache(maxsize=None)
def fetch_algorithms(only_anchors=False, **folders):
    """
    Fetch algorithm from specified algorithms folders.

    By default, uses:
        - integration.algorithms folder
        - default algorithms folder

    :returns: all finded algorithm's classes.
    """

    folders = folders or (
        ALGORITHMS_FOLDER,
        INTEGRATION_FOLDER,
    )

    pool = dict()

    target = (
        script
        for script in chain(
            *[folder.rglob("*.py") for folder in folders]
        )
        if "__" not in script.name
    )

    for script in target:
        spec = importlib.util.spec_from_file_location(
            script.stem, str(script)
        )
        module = importlib.util.module_from_spec(spec)
        sys.modules[module.__name__] = module

        with contextlib.suppress(PackageNotInstalledError):
            spec.loader.exec_module(module)

        iterable = (
            obj
            for obj in inspect.getmembers(module, inspect.isclass)
            if (
                obj[1].__module__ == module.__name__
                and issubclass(obj[1], SearchAlgorithm)
            )
        )

        for name, cls in iterable:
            pool[cls.anchor] = cls

            if not only_anchors:
                pool[name] = cls

                for al in cls.aliases:
                    pool[al] = cls

    return pool


def get_algo(name):
    """Get algorithm class by name"""

    classes = fetch_algorithms()
    cls = classes.get(name)

    if not cls:
        raise SearchAlgorithmNotFoundedError()

    return cls



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Bayesian optimization module."""

import inspect
from typing import Generator
from typing import List
from typing import Optional

import numpy as np
from scipy.stats import norm
from sklearn.ensemble import RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessRegressor

from feijoa.models.configuration import Configuration
from feijoa.search.algorithms import SearchAlgorithm
from feijoa.search.parameters import Categorical
from feijoa.search.parameters import Integer
from feijoa.search.parameters import Real
from feijoa.search.visitors import Randomizer


__all__ = ["BayesianAlgorithm"]


# noinspection PyPep8Naming
class BayesianAlgorithm(SearchAlgorithm):
    """Bayesian optimization algorithm.

    Bayesian optimization is a global optimization
    method for an unknown function (black box) with noise.
    Bayesian optimization applied to hyperparameters
    optimization builds a stochastic model of the mapping
    function from hyperparameter value to an objective
    function applied on the test set. By iteratively
    applying a perspective hyperparameter configuration
    based on the current model and then updating it,
    Bayesian optimization seeks to gather as much
    information as possible about that function and,
    in particular, the location of the optimum. The method
    attempts to balance probing (hyperparameters for which
    the change is least reliably known) and exploitation
    (hyperparameters that are expected to be closest to
    the optimum). In practice, Bayesian optimization
    has shown better results with less computation compared
    to grid search and random search due to the ability
    to judge the quality of experiments even before they
    are performed.

    Args:
        search_space (SearchSpace):
            Search space instance.
        acq_function (str):
            Acquisition function.
            Can be `pi`, `ucb`, `ei` with Gaussian Regressor
            Or `mc` - experimental, `lfboei`, `lfbopi`
        regressor:
            Regression model, must have
                - fit(X, y)
                - predict_proba(X)
                - predict(X)
            methods. For example, you
            can use regressor from sklearn package
            (https://scikit-learn.org/)
        n_warmup (int):
            Warmup points count.

    .. note::
        `lfbopi` and `lfboei` taked from publication:
         https://arxiv.org/abs/2206.13035

    .. note::
        `mc` - experimental function.

    """

    anchor = "bayesian"

    aliases = (
        "BayesianAlgorithm",
        "bayesian",
        "bayes",
    )

    def __init__(
        self,
        search_space,
        *args,
        acq_function="ei",
        regressor=GaussianProcessRegressor,
        n_warmup=5,
        **kwargs,
    ):

        super().__init__(*args, **kwargs)

        self.search_space = search_space

        # build regression model

        self.model = (
            regressor() if inspect.isclass(regressor) else regressor
        )

        # low and high bound of
        # target hyperparameters

        self.bounds = []

        for p in self.search_space:
            if isinstance(p, (Integer, Real)):
                self.bounds.append((p.low, p.high))
            if isinstance(p, Categorical):
                self.bounds.append((0, len(p.choices) - 1))

        self.bounds = np.array(self.bounds)

        # trials pool for fitting

        self.X = np.empty(shape=(0, len(self.search_space)))

        # results pool for fitting

        self.y = np.empty(shape=(0,))

        # setup random generator

        self.random_generator = np.random.RandomState(0)

        # setup acquisition function

        self.acq_function = (
            acq_function
            if isinstance(self.model, GaussianProcessRegressor)
            else "mc"
        )

        # warmup points

        self.n_warmup = n_warmup

        # setup readable name
        self._name = f"Bayesian<{type(self.model).__name__}({self.acq_function})>"

        self._ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:

        if not self._ask_gen:
            # build ask generator
            self._ask_gen = self._ask(n)

        return next(self._ask_gen)

    def _to_feijoa_config(self, x):
        """Build feijoa-format
        configuration from vec x.

        Args:
            x (numpy.array):
                Input x vector for transformation.

        Returns:
            Feijoa-format configuration (dict).

        Raises:
            AnyError: If anything bad happens.

        """

        configuration = {}

        for p, v in zip(self.search_space, x):
            if isinstance(p, Real):
                configuration[p.name] = v

            # make math rounding

            rounded = int(v + (0.5 if v > 0 else -0.5))

            if isinstance(p, Integer):
                configuration[p.name] = rounded

            # pick up needed categorical value

            elif isinstance(p, Categorical):
                configuration[p.name] = p.choices[rounded]

        return configuration

    def _to_gp_config(self, cfg):
        """Build vec from Feijoa-format
        configuration.

        Args:
            cfg (dict | Configuration):
                Input configuration for transformation.

        Returns:
            Values vector (numpy.array).

        Raises:
            AnyError: If anything bad happens.

        """

        vec = []

        for p_name, v in cfg.items():
            p = self.search_space.get(p_name)

            if isinstance(p, (Integer, Real)):
                vec.append(v)
            if isinstance(p, Categorical):
                index = p.choices.index(v)
                vec.append(index)

        return vec

    def _ask(self, n: int) -> Generator:
        """Main ask generator."""

        # make some warmup configurations

        randomizer = Randomizer()
        random_samples = [
            Configuration(
                {
                    p.name: p.accept(randomizer)
                    for p in self.search_space
                },
                requestor=self.name,
            )
            for _ in range(self.n_warmup)
        ]

        yield random_samples

        self.model.fit(self.X, self.y)

        while True:
            x = self.opt_acquisition(n)
            configurations = list()
            for c in x:
                configurations.append(
                    Configuration(
                        self._to_feijoa_config(c), requestor=self.name
                    )
                )
            yield configurations
            self.model.fit(self.X, self.y)

    def acquisition(self, X_samples):
        """Acquisition function for bayesian optimization"""

        yhat = self.model.predict(self.X)
        best = min(yhat)

        if self.acq_function == "mc":
            # experimental, research only

            mean = self.model.predict(X_samples)
            negative = np.where(mean - best < 0)
            mean[negative] = 0
            return mean - best

        if self.acq_function == "lfboei":
            # https://arxiv.org/abs/2206.13035

            tau = np.quantile(self.y, q=0.33)
            classified = np.greater_equal(self.y, tau)
            x1, z1 = self.X[classified], classified[classified]
            x0, z0 = self.X, np.zeros_like(classified)
            w1 = (tau - self.y)[classified]
            w1 = w1 / np.mean(w1)
            w0 = 1 - z0
            x = np.concatenate([x1, x0], axis=0)
            z = np.concatenate([z1, z0], axis=0)
            s1 = x1.shape[0]
            s0 = x0.shape[0]
            w = np.concatenate(
                [w1 * (s1 + s0) / s1, w0 * (s1 + s0) / s0], axis=0
            )
            w = w / np.mean(w)

            clf = RandomForestClassifier(n_jobs=-1)
            clf.fit(x, z, sample_weight=w)
            y_pred = clf.predict_proba(X_samples)[:, 1]
            return y_pred

        if self.acq_function == "lfbopi":
            # https://arxiv.org/abs/2206.13035

            tau = np.quantile(self.y, q=0.33)
            classified = np.greater_equal(self.y, tau)
            clf = RandomForestClassifier(n_jobs=-1)
            clf.fit(self.X, classified)
            y_pred = clf.predict_proba(X_samples)[:, 1]
            return y_pred

        mean, std = self.model.predict(X_samples, return_std=True)
        kappa = 2.5

        # regular acquisition functions.

        if self.acq_function == "poi":
            probs = norm.cdf((mean - best) / (std + 1e-9))
            return probs
        elif self.acq_function == "ei":
            a = mean - best
            z = a / std
            return a * norm.cdf(z) + std * norm.pdf(z)
        elif self.acq_function == "ucb":
            return mean + kappa * std

    def opt_acquisition(self, n: int):
        """Optimize acquisition function."""

        n_warmup = 10000

        x_tries = self.random_generator.uniform(
            self.bounds[:, 0],
            self.bounds[:, 1],
            size=(n_warmup, self.bounds.shape[0]),
        )

        scores = self.acquisition(x_tries)
        minima_x = x_tries[scores.argsort()[:n]]

        return minima_x

    def tell(self, config, result):
        """Tell configuration's result."""

        vec = np.array(self._to_gp_config(config))

        self.X = np.concatenate([self.X, vec.reshape(1, -1)])
        self.y = np.concatenate([self.y, [result]])



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import abc
from typing import List
from typing import Optional

from feijoa.models import Experiment
from feijoa.search.space import SearchSpace


__all__ = ["Storage"]


class Storage(metaclass=abc.ABCMeta):
    def insert_job(self, job):
        """Insert job into storage.

        Args:
            job (Job):
                Job instance.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def is_job_name_exists(self, name):
        """Find job in storage.

        Args:
            name (str):
                Job name.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_job_id_by_name(self, name) -> Optional[int]:
        """Get job by id in storage.

        Args:
            name (str):
                Name of job.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def insert_experiment(self, experiment):
        """Insert experiment into storage.

        Args:
            experiment (Experiment):
                Experiment instance to insert.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_experiment(self, job_id, experiment_id):
        """Get experiment from job.

        Args:
            job_id (int):
                Job index.
            experiment_id (int):
                Experiment index.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_experiments_by_job_id(self, job_id) -> List[Experiment]:
        """Get experiments from job

        Args:
            job_id:
                Index of job in storage.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_experiments_count(self, job_id) -> int:
        """Get experiments count in specified job.

        Args:
            job_id (int):
                Job id.

        Raises:
            AnyError: If anything bad happens.

        """

        pass

    def best_experiment(self, job) -> Optional[Experiment]:
        """Get best experiment from job.

        Args:
            job (Job):
                Job instance.

        Raises:
            AnyError: If anything bad happens.

        """

        experiments = self.get_experiments_by_job_id(job)

        if not experiments:
            return None

        return min(experiments, key=lambda x: x.objective_result)  # type: ignore

    def top_experiments(self, job_id, n) -> List[Experiment]:
        """Get top experiments from job.
        Experiments is sorted by objective values.

        Args:
            job_id (int):
                Job index.

        Raises:
            AnyError: If anything bad happens.

        """

        experiments = self.get_experiments_by_job_id(job_id)

        experiments.sort(key=lambda x: x.objective_result)  # type: ignore
        return experiments[: min(len(experiments), n)]

    def get_search_space_by_job_id(self, job_id) -> SearchSpace:
        """Get search space from existed job.

        Args:
            job_id (int):
                Job index.

        Raises:
            AnyError: If anything bad happens.

        """

        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def jobs(self):
        """Pick up jobs from storage."""

        raise NotImplementedError()

    @property
    def jobs_count(self):
        """Get jobs count from storage."""

        return len(self.jobs)

    @property
    @abc.abstractmethod
    def version(self):
        """Get version of database schema."""

        raise NotImplementedError()



# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Meta algorithm base class module."""

import abc
import logging
from typing import Generator
from typing import List
from typing import Optional

from feijoa.models.configuration import Configuration
from feijoa.search.algorithms import SearchAlgorithm
from feijoa.search.seed import SeedAlgorithm


log = logging.getLogger(__name__)


class MetaSearchAlgorithm(SearchAlgorithm):
    """Meta-search algorithm.

    Used for pick up algorithms.

    Raises:
        AnyError: If anything bad happens.

    """

    def __init__(self, *algorithms, **kwargs):
        super().__init__(*algorithms, **kwargs)
        self.algorithms = list(algorithms)
        self.ask_gen = None

    def ask(self, n: int = 1) -> Optional[List[Configuration]]:
        if not self.ask_gen:
            self.ask_gen = self._ask(n)
        return next(self.ask_gen)

    def _ask(self, n: int) -> Generator:
        prev_picked = ""

        # pick up all seed's algorithms

        seed_algorithms = [
            a for a in self.algorithms if isinstance(a, SeedAlgorithm)
        ]

        # remove all seed's algorithms

        for algo in seed_algorithms:
            self.remove_algorithm(algo)

        if seed_algorithms:
            log.debug("Let's measure seed configurations!")
            for algo in seed_algorithms:
                configs = algo.ask(n)
                yield configs

        while True:
            for algo in self.order:
                if not prev_picked or prev_picked != algo.name:
                    log.debug(f"Pick {algo.name}")
                configs = algo.ask(n)
                yield configs
                prev_picked = algo.name

    def tell(self, config, result):
        """Tell results to all search algorithms"""

        for algo in self.algorithms:
            algo.tell(config, result)

    def add_algorithm(self, algo: SearchAlgorithm):
        """Append algorithm to algorithms list.

        Args:
            algo (SearchAlgorithm):
                search algorithm instance.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

        self.algorithms.append(algo)

    @abc.abstractmethod
    def remove_algorithm(self, algo: SearchAlgorithm):
        """Remove algorithm from inner techniques
        pool in meta-technique.

        Args:
            algo (SearchAlgorithm):
                search algorithm instance.

        Returns:
            None

        Raises:
            AnyError: If anything bad happens.

        """

    @property
    @abc.abstractmethod
    def order(self):
        """Get order for algorithms."""

        raise NotImplementedError()
